package se.uu.bench;

import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.atomic.AtomicLongArray;

abstract public class Benchmark{


    AtomicLongArray paddedStoppedFlag = new AtomicLongArray(33);
    protected volatile boolean shallStart = false;

    private boolean timeLimited;
    private int nrOfWarmUpRuns;
    private double warmUpRunSeconds;
    private int nrOfMeasurmentRuns;
    private double measurmentRunSeconds;
    private ResultReportType reportType;
    protected int nrOfThreads;
    protected CyclicBarrier waitBarrier = null;
    protected String setType;
    protected long globalAvoidOptOutSum = 0;

    public enum ResultReportType {
        ALL,AVERAGE_MIN_MAX,AVERAGE,AVERAGE_THROUGHPUT
    }

    protected abstract class WorkerThread extends Thread{
        protected long measurment = 0;
        protected long avoidOptOutSum = 0;

        public long getMeasurment(){
            return measurment;
        }

        public long getAvoidOptOutSum(){
            return measurment;
        }

        public abstract void setUp();
        
    }

    protected WorkerThread[] threads;

    protected interface AbstractElem extends Comparable<AbstractElem>{
        public int compareTo(AbstractElem o);
        public String toString();
        public int hashCode();
        public int getKey();
        public void setKey(int key);
    }

    protected final class Elem implements AbstractElem{
        private int key;
        public Elem(int key){
            this.key = key;
        }
        public int compareTo(AbstractElem o){
            return key - o.getKey();
        }
        public String toString(){
            return new Integer(key).toString();
        }
        public int hashCode() {
            return key * 0x9e3775cd;
        }

        public int getKey(){
            return key;
        }

        public void setKey(int key){
            this.key = key;
        }
        
    }

    //We do not want any false sharing in this elements
    protected final class WorkingElem implements AbstractElem{
        private int key[] = new int[65];
        public int compareTo(AbstractElem o){
            return key[32] - o.getKey();
        }
        public String toString(){
            return new Integer(key[32]).toString();
        }
        public int hashCode() {
            return key[32] * 0x9e3775cd;
        }
        
        public int getKey(){
            return key[32];
        }
        
        public void setKey(int key){
            this.key[32] = key;
        }
    };

    abstract protected void setUp() throws Exception;


    public long tearDown() throws Exception{
        long sum = 0;
        for(int i = 0; i < nrOfThreads; i++){
            threads[i].join();
            sum = sum + threads[i].getMeasurment();
            globalAvoidOptOutSum = globalAvoidOptOutSum + threads[i].getAvoidOptOutSum();
            threads[i] = null;
        }
        paddedStoppedFlag.set(16, 0);
        shallStart = false;
        Thread.sleep(1000);
        System.err.println("#"+globalAvoidOptOutSum+"#");
        return sum;
    }

    public boolean stopped(){       
        return paddedStoppedFlag.get(16) == 1;
    }

    public boolean shallStart(){
        return shallStart;
    }

    public Benchmark(boolean timeLimited,
                     int nrOfWarmUpRuns,
                     double warmUpRunSeconds,
                     int nrOfMeasurmentRuns,
                     double measurmentRunSeconds,
                     ResultReportType reportType,
                     int nrOfThreads,
                     String setType){
        this.nrOfWarmUpRuns = nrOfWarmUpRuns;
        this.warmUpRunSeconds = warmUpRunSeconds;
        this.nrOfMeasurmentRuns = nrOfMeasurmentRuns;
        this.measurmentRunSeconds = measurmentRunSeconds;
        this.reportType = reportType;
        this.nrOfThreads = nrOfThreads;
        this.timeLimited = timeLimited;
        this.setType = setType;
    }

    public void start() throws Exception{
        long warmUpRunMillis = (long)(warmUpRunSeconds * 1000);
        long measurmentRunMillis = (long)(measurmentRunSeconds * 1000);
        //Warm up
        if(!timeLimited){
            waitBarrier = new CyclicBarrier(nrOfThreads + 1);
        }
        for(int i = 0; i < nrOfWarmUpRuns; i++){
            System.err.println("=> Seting up warmup run");
            setUp();
            System.err.println("== Benchmark setup completed");
            Thread.sleep(100);
            System.err.println("=> Starting warm up run");
            long started = System.nanoTime();
            shallStart = true;
            if(timeLimited){
                Thread.sleep(warmUpRunMillis);
            }else{
                waitBarrier.await();
            }
            paddedStoppedFlag.set(16, 1);
            long ended = System.nanoTime();
            long nrOfOpsPerformed = tearDown();
            double throuput = new Long(nrOfOpsPerformed).doubleValue() / new Long((ended - started)).doubleValue();
            System.err.println("== Warmup run completed " + throuput + " ops/nanosecond");
        }
        
        long[] benchmarkResults = new long[nrOfMeasurmentRuns];
        long[] benchmarkTimes = new long[nrOfMeasurmentRuns];
        //Benchmark
        for(int i = 0; i < nrOfMeasurmentRuns; i++){
            System.err.println("=> Seting up benchmark run");
            setUp();
            System.err.println("== Benchmark setup completed");
            Thread.sleep(100);
            System.err.println("=> Starting benchmark run");
            long started = System.nanoTime();
            shallStart = true;
            if(timeLimited){
                Thread.sleep(measurmentRunMillis);
            }else{
                waitBarrier.await();
            }
            paddedStoppedFlag.set(16, 1);
            long ended = System.nanoTime();
            long nrOfOpsPerformed = tearDown();
            benchmarkResults[i] = nrOfOpsPerformed;
            benchmarkTimes[i] = ended - started;
            double throuput = new Long(nrOfOpsPerformed).doubleValue() / new Long((ended - started)).doubleValue();
            System.err.println("== Benchmark run completed " + throuput + " ops/nanosecond");
        }
        System.out.print(setType + " " + nrOfThreads + " ");
        if(reportType == ResultReportType.ALL){          
            for(int i = 0; i < nrOfMeasurmentRuns; i++){
                System.out.print(benchmarkTimes[i]);
                System.out.print(" ");
                System.out.print(benchmarkResults[i]);
                if(i < (nrOfMeasurmentRuns-1)){
                    System.out.print(" ");
                }
            }
            System.out.print("\n");
        }else if(reportType == ResultReportType.AVERAGE_MIN_MAX){
            long sumMeasurments = 0;
            long sumTimes = 0;

            long minMeasurment = Long.MAX_VALUE;
            int minMeasurmentIndex = -1;
            long maxMeasurment = Long.MIN_VALUE;
            int maxMeasurmentIndex = -1;

            long minTime = Long.MAX_VALUE;
            int minTimeIndex = -1;
            long maxTime = Long.MIN_VALUE;
            int maxTimeIndex = -1;
            for(int i = 0; i < nrOfMeasurmentRuns; i++){
                if(benchmarkTimes[i] < minTime){
                    minTime = benchmarkTimes[i];
                    minTimeIndex = i;
                }
                if(benchmarkTimes[i] > maxTime){
                    maxTime = benchmarkTimes[i];
                    maxTimeIndex = i;
                }
                if(benchmarkResults[i] < minMeasurment){
                    minMeasurment = benchmarkResults[i];
                    minMeasurmentIndex = i;
                }
                if(benchmarkResults[i] > maxMeasurment){
                    maxMeasurment = benchmarkResults[i];
                    maxMeasurmentIndex = i;
                }
                sumTimes = sumTimes + benchmarkTimes[i];
                sumMeasurments = sumMeasurments + benchmarkResults[i];               
            }
            System.out.print(((double)sumTimes)/nrOfMeasurmentRuns + " ");
            System.out.print(((double)sumMeasurments)/nrOfMeasurmentRuns + " ");
            System.out.print(benchmarkTimes[minTimeIndex] + " " + benchmarkResults[minTimeIndex] + " ");
            System.out.print(benchmarkTimes[maxTimeIndex] + " " + benchmarkResults[maxTimeIndex] + " ");
            System.out.print(benchmarkTimes[minMeasurmentIndex] + " " + benchmarkResults[minMeasurmentIndex] + " ");
            System.out.print(benchmarkTimes[maxMeasurmentIndex] + " " + benchmarkResults[maxMeasurmentIndex]);
        }else if(reportType == ResultReportType.AVERAGE){
            long sumMeasurments = 0;
            long sumTimes = 0;
            for(int i = 0; i < nrOfMeasurmentRuns; i++){
                sumTimes = sumTimes + benchmarkTimes[i];
                sumMeasurments = sumMeasurments + benchmarkResults[i];               
            }
            System.out.print(((double)sumTimes)/nrOfMeasurmentRuns + " ");
            System.out.println(((double)sumMeasurments)/nrOfMeasurmentRuns);
        }else if(reportType == ResultReportType.AVERAGE_THROUGHPUT){
            long sumMeasurments = 0;
            long sumTimes = 0;
            for(int i = 0; i < nrOfMeasurmentRuns; i++){
                sumTimes = sumTimes + benchmarkTimes[i];
                sumMeasurments = sumMeasurments + benchmarkResults[i];               
            }
            System.out.println(((double)sumMeasurments)/((double)sumTimes));
        }
    }

}
